"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const http_transport_circuit_breaker_1 = require("@bbc/http-transport-circuit-breaker");
const _ = require("lodash");
const DEFAULT_MAX_FAILURES = 100;
const DEFAULT_RESET_TIMEOUT = 10000;
const DEFAULT_TIMEOUT = 0x7FFFFFFF;
function isFailure(err) {
    return err && err.statusCode >= 500;
}
function getExecutor(client) {
    return function (...args) {
        const method = args[0];
        const rest = args.slice(1, args.length - 1);
        const cb = args[args.length - 1];
        return client[method].apply(client, rest)
            .then((result) => {
            cb(null, result);
        })
            .catch((err) => {
            cb(err);
        });
    };
}
class CircuitBreaker {
    constructor(client, params) {
        const maxFailures = _.get(params, 'circuitbreaker.maxFailures', DEFAULT_MAX_FAILURES);
        const resetTimeout = _.get(params, 'circuitbreaker.resetTimeout', DEFAULT_RESET_TIMEOUT);
        const opts = {
            isFailure,
            maxFailures,
            resetTimeout,
            timeout: DEFAULT_TIMEOUT,
            openErrMsg: `[${params.name}] Circuit breaker is open`
        };
        const breaker = http_transport_circuit_breaker_1.createBreaker({ execute: getExecutor(client) }, opts);
        if (params.stats) {
            breaker.on('open', () => {
                params.stats.increment(`${params.name}.circuit_breaker.open`);
            });
        }
        this.executor = async (...args) => {
            return new Promise((resolve, reject) => {
                breaker.run(...args, (err, result) => {
                    if (err) {
                        reject(err);
                        return;
                    }
                    resolve(result);
                });
            });
        };
    }
    async get(uri, opts) {
        return this.executor('get', ...arguments);
    }
    async patch(uri, body, opts) {
        return this.executor('patch', ...arguments);
    }
    async post(uri, body, opts) {
        return this.executor('post', ...arguments);
    }
    async put(uri, body, opts) {
        return this.executor('put', ...arguments);
    }
    async head(uri, opts) {
        return this.executor('head', ...arguments);
    }
    async delete(uri, opts) {
        return this.executor('delete', ...arguments);
    }
}
exports.CircuitBreaker = CircuitBreaker;
