"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const httpTransport = require("@bbc/http-transport");
const http_transport_cache_1 = require("@bbc/http-transport-cache");
const rateLimit = require("@bbc/http-transport-rate-limiter");
const collapsing = require("@bbc/http-transport-request-collapse");
const stats = require("@bbc/http-transport-statsd");
const toError = require("@bbc/http-transport-to-error");
const request = require("request");
const memoryCache = require("../caching/memory");
const logger = httpTransport.logger;
const defaultRetries = 1;
const defaultDelay = 100;
exports.defaultRequestOptions = Object.freeze({
    json: true,
    timeout: 2000,
    time: true,
    gzip: true,
    forever: true,
    proxy: undefined,
    agentOptions: {
        maxSockets: 1000
    }
});
function getRetries(params) {
    return params.retries === undefined ? defaultRetries : params.retries;
}
function getRetryDelay(params) {
    return params.retryDelay === undefined ? defaultDelay : params.retryDelay;
}
function configureMemoryCache(builder, params) {
    const cache = memoryCache.createCache(params.memoryCache);
    const cacheOpts = {
        name: `${params.name}memory`,
        varyOn: params.varyOn
    };
    builder.use(http_transport_cache_1.maxAge(cache, cacheOpts));
    builder.use(http_transport_cache_1.staleIfError(cache, cacheOpts));
    configureStatsEvents({
        client: params.stats,
        namespace: params.name,
        statName: 'memory_cache',
        eventName: cacheOpts.name
    });
}
function configureStatsEvents(params) {
    const { client, namespace, statName, eventName } = params;
    if (client) {
        http_transport_cache_1.events.on(`cache.${eventName}.hit`, () => {
            client.increment(`${namespace}.${statName}.hits`, 1, 0.05);
        });
        http_transport_cache_1.events.on(`cache.${eventName}.miss`, () => {
            client.increment(`${namespace}.${statName}.misses`, 1, 0.05);
        });
        http_transport_cache_1.events.on(`cache.${eventName}.error`, () => {
            client.increment(`${namespace}.${statName}.errors`);
        });
        http_transport_cache_1.events.on(`cache.${eventName}.timeout`, () => {
            client.increment(`${namespace}.${statName}.timeouts`);
        });
        http_transport_cache_1.events.on(`cache.${eventName}.stale`, () => {
            client.increment(`${namespace}.${statName}.stale`, 1, 0.05);
        });
        http_transport_cache_1.events.on(`cache.${eventName}.revalidate.error`, () => {
            client.increment(`${namespace}.${statName}.revalidate.error`, 1, 0.05);
        });
        http_transport_cache_1.events.on(`cache.${eventName}.revalidate`, () => {
            client.increment(`${namespace}.${statName}.revalidate`, 1, 0.05);
        });
    }
}
function configureExternalCache(builder, params) {
    const cache = params.externalCache.cache;
    const cacheOpts = {
        name: params.name,
        varyOn: params.varyOn
    };
    if (params.externalCache.timeout) {
        cacheOpts.timeout = params.externalCache.timeout;
        cacheOpts.ignoreCacheErrors = true;
    }
    builder.use(http_transport_cache_1.maxAge(cache, cacheOpts));
    builder.use(http_transport_cache_1.staleIfError(cache, cacheOpts));
    configureStatsEvents({
        client: params.stats,
        namespace: params.name,
        statName: 'cache',
        eventName: cacheOpts.name
    });
}
function configureRequest(params) {
    if (params.httpClient) {
        return params.httpClient;
    }
    const defaults = Object.assign({}, exports.defaultRequestOptions);
    return new httpTransport.defaultTransport(request.defaults(defaults));
}
function configureCollapsing(builder, params) {
    builder.use(collapsing.middleware({
        eventName: params.name,
        collapsingWindow: params.collapsing.window
    }));
    if (params.stats) {
        collapsing.events.on(`collapsed-${params.name}`, () => {
            params.stats.increment(`${params.name}.collapsing.collapsed`);
        });
    }
}
function configureClient(params) {
    const transport = configureRequest(params);
    const builder = httpTransport
        .createBuilder(transport)
        .retries(getRetries(params))
        .retryDelay(getRetryDelay(params))
        .use(toError());
    if (params.rateLimit && params.rateLimitInterval) {
        builder.use(rateLimit(params.rateLimit, params.rateLimitInterval));
    }
    if (params.userAgent !== undefined) {
        builder.userAgent(params.userAgent);
    }
    if (params.collapsing) {
        configureCollapsing(builder, params);
    }
    if (params.memoryCache) {
        configureMemoryCache(builder, params);
    }
    if (params.externalCache) {
        configureExternalCache(builder, params);
    }
    if (params.logger) {
        builder.use(logger(params.logger));
    }
    if (params.stats) {
        builder.use(stats(params.stats, params.name));
    }
    return builder.createClient();
}
exports.configureClient = configureClient;
